import time

from maplib import Mapping
import polars as pl
import os
import pathlib
from pathlib import Path
import pytest

pl.Config.set_fmt_str_lengths(300)

PATH_HERE = pathlib.Path(__file__).parent
GTFS_PATH = PATH_HERE / "testdata" / "gtfs"


@pytest.fixture(scope="function")
def mapping() -> Mapping:
    time_start = time.time()
    mapping_prefixes = """
    @prefix rr:<http://www.w3.org/ns/r2rml#>.
    @prefix foaf:<http://xmlns.com/foaf/0.1/>.
    @prefix xsd:<http://www.w3.org/2001/XMLSchema#>.
    @prefix rdfs:<http://www.w3.org/2000/01/rdf-schema#>.
    @prefix dc:<http://purl.org/dc/elements/1.1/>.
    @prefix rev:<http://purl.org/stuff/rev#>.
    @prefix gtfs:<http://vocab.gtfs.org/terms#>.
    @prefix geo:<http://www.w3.org/2003/01/geo/wgs84_pos#>.
    @prefix schema:<http://schema.org/>.
    @prefix dct:<http://purl.org/dc/terms/>.
    @prefix rml:<http://semweb.mmlab.be/ns/rml#>.
    @prefix ql:<http://semweb.mmlab.be/ns/ql#>.
    @prefix rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    @prefix t:<https://github.com/magbak/maplib/benchmark/template#>.
    """

    i = 5

    path = GTFS_PATH / "datasets" / "csv" / f"{i}"

    frequencies = pl.scan_csv(path / "FREQUENCIES.csv")
    calendar_dates = pl.scan_csv(path / "CALENDAR_DATES.csv")
    trips = pl.scan_csv(path / "TRIPS.csv")
    stop_times = pl.scan_csv(path / "STOP_TIMES.csv")
    stops = pl.scan_csv(path / "STOPS.csv")
    routes = pl.scan_csv(path / "ROUTES.csv")
    feed_info = pl.scan_csv(path / "FEED_INFO.csv")
    agency = pl.scan_csv(path / "AGENCY.csv")
    calendar = pl.scan_csv(path / "CALENDAR.csv")
    shapes = pl.scan_csv(path / "SHAPES.csv")

    TRANS_LINKED_ES = "http://transport.linkeddata.es/"
    STOPTIMES_PREFIX = TRANS_LINKED_ES + "madrid/metro/stoptimes/"
    PICKUP_TYPE_PREFIX = TRANS_LINKED_ES + "resource/PickupType/"
    DROPOFF_TYPE_PREFIX = TRANS_LINKED_ES + "resource/DropOffType/"
    TRIP_ID_PREFIX = TRANS_LINKED_ES + "madrid/metro/trips/"
    ROUTE_ID_PREFIX = TRANS_LINKED_ES + "madrid/metro/routes/"
    ROUTE_TYPE_PREFIX = TRANS_LINKED_ES + "resource/RouteType/"
    STOP_ID_PREFIX = TRANS_LINKED_ES + "madrid/metro/stops/"
    SERVICE_ID_PREFIX = TRANS_LINKED_ES + "madrid/metro/services/"
    AGENCY_ID_PREFIX = TRANS_LINKED_ES + "madrid/agency/"
    SHAPE_ID_PREFIX = TRANS_LINKED_ES + "madrid/metro/shape/"
    SHAPE_POINT_PREFIX = TRANS_LINKED_ES + "madrid/metro/shape_point/"
    FREQUENCY_ID_PREFIX = TRANS_LINKED_ES + "madrid/metro/frequency/"
    WHEELCHAIR_ACCESSIBLE_PREFIX = TRANS_LINKED_ES + "resource/WheelchairBoardingStatus/"
    LOCATION_TYPE_PREFIX = TRANS_LINKED_ES + "resource/LocationType/"
    CALENDAR_DATE_RULES_PREFIX = TRANS_LINKED_ES + "madrid/metro/calendar_date_rule/"
    CALENDAR_RULES_PREFIX = TRANS_LINKED_ES + "madrid/metro/calendar_rules/"
    FEED_PUBLISHER_PREFIX = TRANS_LINKED_ES + "madrid/metro/feed/"

    stop_times_mut = stop_times.with_columns([
        pl.col("stop_sequence").cast(pl.Int32),
        (STOPTIMES_PREFIX + pl.col("trip_id") + '-' + pl.col("stop_id") + "-" + pl.col("arrival_time")).alias(
            "subject"),
        (PICKUP_TYPE_PREFIX + pl.col("pickup_type")).alias("pickup_type"),
        (DROPOFF_TYPE_PREFIX + pl.col("drop_off_type")).alias("drop_off_type"),
        (TRIP_ID_PREFIX + pl.col("trip_id")).alias("trip_id"),
        (STOP_ID_PREFIX + pl.col("stop_id")).alias("stop_id")
    ]).collect()

    stop_times_mapping = mapping_prefixes + """
    t:StopTimes [xsd:anyURI ?trip_id, ?arrival_time, ?departure_time, xsd:anyURI ?stop_id, ??stop_sequence, 
                 ?stop_headsign, ?pickup_type, ?drop_off_type, ?shape_dist_traveled,  xsd:anyURI ?subject ] :: { 
      ottr:Triple(?subject,rdf:type, gtfs:StopTime),
      ottr:Triple(?subject,gtfs:trip,?trip_id),
      ottr:Triple(?subject,gtfs:arrivalTime,?arrival_time),
      ottr:Triple(?subject,gtfs:departureTime,?departure_time),
      ottr:Triple(?subject,gtfs:stop,?stop_id),
      ottr:Triple(?subject,gtfs:stopSequence,?stop_sequence),
      ottr:Triple(?subject,gtfs:headsign,?stop_headsign),
      ottr:Triple(?subject,gtfs:pickupType,?pickup_type),
      ottr:Triple(?subject,gtfs:dropoffType,?drop_off_type),
      ottr:Triple(?subject,gtfs:distanceTraveled,?shape_dist_traveled)
    }.
    """

    trips_mut = trips.with_columns([
        (TRIP_ID_PREFIX + pl.col("trip_id")).alias("trip_id"),
        (WHEELCHAIR_ACCESSIBLE_PREFIX + pl.col("wheelchair_accessible")).alias("wheelchair_accessible"),
        (SERVICE_ID_PREFIX + pl.col("service_id")).alias("service_id"),
        (ROUTE_ID_PREFIX + pl.col("route_id")).alias("route_id"),
        (SHAPE_ID_PREFIX + pl.col("shape_id")).alias("shape_id")
    ]).collect()

    trips_mapping = mapping_prefixes + """
    t:Trips [xsd:anyURI ?route_id, xsd:anyURI ?service_id,  xsd:anyURI ?trip_id, ?trip_headsign, ?trip_short_name, 
             xsd:anyURI ?direction_id, xsd:anyURI ?block_id, xsd:anyURI ?shape_id, ?wheelchair_accessible ] :: {
      ottr:Triple(?trip_id, rdf:type, gtfs:Trip) ,
      ottr:Triple(?trip_id, gtfs:route, ?route_id) ,
      ottr:Triple(?trip_id, gtfs:service, ?service_id) ,
      ottr:Triple(?trip_id, gtfs:headsign, ?trip_headsign) ,
      ottr:Triple(?trip_id, gtfs:shortName, ?trip_short_name) ,
      ottr:Triple(?trip_id, gtfs:direction, ?direction_id) ,
      ottr:Triple(?trip_id, gtfs:block, ?block_id) ,
      ottr:Triple(?trip_id, gtfs:shape, ?shape_id) ,
      ottr:Triple(?trip_id, gtfs:wheelchairAccessible, ?wheelchair_accessible)
    } . """

    routes_mut = routes.with_columns([
        (ROUTE_ID_PREFIX + pl.col("route_id")).alias("route_id"),
        (ROUTE_TYPE_PREFIX + pl.col("route_type")).alias("route_type"),
        (AGENCY_ID_PREFIX + pl.col("agency_id")).alias("agency_id")
    ]).collect()

    routes_mapping = mapping_prefixes + """
    t:Routes [ xsd:anyURI ?route_id, xsd:anyURI ?agency_id, ?route_short_name, ?route_long_name, ?route_desc, 
               ?route_type, xsd:anyURI ?route_url, ?route_color, ?route_text_color ] :: {
      ottr:Triple(?route_id, rdf:type, gtfs:Route) ,
      ottr:Triple(?route_id, gtfs:agency, ?agency_id) ,
      ottr:Triple(?route_id, gtfs:shortName, ?route_short_name) ,
      ottr:Triple(?route_id, gtfs:longName, ?route_long_name) ,
      ottr:Triple(?route_id, dct:description, ?route_desc) ,
      ottr:Triple(?route_id, gtfs:routeType, ?route_type) ,
      ottr:Triple(?route_id, gtfs:routeUrl, ?route_url) ,
      ottr:Triple(?route_id, gtfs:color, ?route_color) ,
      ottr:Triple(?route_id, gtfs:textColor, ?route_text_color)
    } . 
    """

    agency_mut = agency.with_columns([
        (AGENCY_ID_PREFIX + pl.col("agency_id")).alias("agency_id"),
        pl.col("agency_name").cast(pl.String)
    ]).collect()

    agency_mapping = mapping_prefixes + """
    t:Agency [ xsd:anyURI ?agency_id, ?agency_name, xsd:anyURI ?agency_url, ?agency_timezone, ?agency_lang, 
               ?agency_phone, xsd:anyURI ?agency_fare_url ] :: {
      ottr:Triple(?agency_id,rdf:type, gtfs:Agency) ,
      ottr:Triple(?agency_id,foaf:name, ?agency_name) ,
      ottr:Triple(?agency_id,foaf:page, ?agency_url) ,
      ottr:Triple(?agency_id,gtfs:timeZone, ?agency_timezone) ,
      ottr:Triple(?agency_id,dct:language,?agency_lang) ,
      ottr:Triple(?agency_id,foaf:phone,?agency_phone) ,
      ottr:Triple(?agency_id,gtfs:fareUrl,?agency_fare_url)
    } ."""

    stops_mut = stops.with_columns([
        (STOP_ID_PREFIX + pl.col("stop_id")).alias("stop_id"),
        (WHEELCHAIR_ACCESSIBLE_PREFIX + pl.col("wheelchair_boarding")).alias("wheelchair_boarding"),
        (LOCATION_TYPE_PREFIX + pl.col("location_type")).alias("location_type"),
        (STOP_ID_PREFIX + pl.col("parent_station")).alias("parent_station"),
        pl.col("stop_lat").cast(pl.Float64),
        pl.col("stop_lon").cast(pl.Float64)
    ]).collect()

    stops_mapping = mapping_prefixes + """
    t:Stops [ xsd:anyURI ?stop_id, ?stop_code, ?stop_name, ?stop_desc, ?stop_lat, ?stop_lon, ??zone_id, 
              xsd:anyURI ?stop_url, xsd:anyURI ?location_type, ? xsd:anyURI ?parent_station, ??stop_timezone, 
              ?wheelchair_boarding ] :: {
      ottr:Triple(?stop_id, rdf:type, gtfs:Stop) ,
      ottr:Triple(?stop_id, gtfs:code, ?stop_code) ,
      ottr:Triple(?stop_id, foaf:name, ?stop_name) ,
      ottr:Triple(?stop_id, dct:description, ?stop_desc) ,
      ottr:Triple(?stop_id, geo:lat, ?stop_lat) ,
      ottr:Triple(?stop_id, geo:long, ?stop_lon) ,
      ottr:Triple(?stop_id, gtfs:zone, ?zone_id) ,
      ottr:Triple(?stop_id, foaf:page, ?stop_url) ,
      ottr:Triple(?stop_id, gtfs:locationType, ?location_type) ,
      ottr:Triple(?stop_id, gtfs:parentStation, ?parent_station) ,
      ottr:Triple(?stop_id, gtfs:timeZone, ?stop_timezone) ,
      ottr:Triple(?stop_id, gtfs:wheelchairAccessible, ?wheelchair_boarding)
    } . 
    """
    #  ottr:Triple(?stop_id, dct:identifier, ?stop_id) ,

    calendar_dates_mut = calendar_dates.with_columns([
        (CALENDAR_DATE_RULES_PREFIX + pl.col("service_id") + "-" + pl.col("date")).alias("subject"),
        (SERVICE_ID_PREFIX + pl.col("service_id")).alias("service_id"),
        (pl.col("date").str.strptime(pl.Date, "%Y-%m-%d"))
    ]).select(["subject", "service_id", "date", "exception_type"
               ]).collect()

    calendar_dates_mapping = mapping_prefixes + """
    t:CalendarDateRules [?date, ?exception_type,  xsd:anyURI ?subject, xsd:anyURI ?service_id ] :: {
      ottr:Triple(?service_id, rdf:type, gtfs:Service),
      ottr:Triple(?service_id, gtfs:serviceRule, ?subject) ,
      ottr:Triple(?subject,rdf:type,gtfs:CalendarDateRule) ,
      ottr:Triple(?subject,dct:date,?date) ,
      ottr:Triple(?subject,gtfs:dateAddition,?exception_type)
    } . """

    tobools = []
    for d in ["monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"]:
        tobools.append(pl.col(d).cast(pl.Boolean))

    calendar_mut = calendar.with_columns([
                                             (CALENDAR_RULES_PREFIX + pl.col("service_id")).alias("subject"),
                                             (SERVICE_ID_PREFIX + pl.col("service_id")).alias("service_id"),
                                             (pl.col("start_date").str.strptime(pl.Date, "%Y-%m-%d")),
                                             (pl.col("end_date").str.strptime(pl.Date, "%Y-%m-%d")),
                                         ] + tobools).collect()

    calendar_mapping = mapping_prefixes + """
    t:CalendarRules [ xsd:anyURI ?subject, xsd:anyURI ?service_id, ?monday, ?tuesday, ?wednesday, ?thursday, ?friday, 
                      ?saturday, ?sunday, ?start_date, ?end_date ] :: {
      ottr:Triple(?service_id, rdf:type, gtfs:Service),
      ottr:Triple(?service_id, gtfs:serviceRule, ?subject) ,
      ottr:Triple(?subject,rdf:type, gtfs:CalendarRule) ,
      ottr:Triple(?subject,gtfs:monday, ?monday) ,
      ottr:Triple(?subject,gtfs:tuesday, ?tuesday) ,
      ottr:Triple(?subject,gtfs:wednesday, ?wednesday) ,
      ottr:Triple(?subject,gtfs:thursday, ?thursday) ,
      ottr:Triple(?subject,gtfs:friday, ?friday) ,
      ottr:Triple(?subject,gtfs:saturday, ?saturday) ,
      ottr:Triple(?subject,gtfs:sunday, ?sunday) ,
      ottr:Triple(?subject,schema:startDate, ?start_date) ,
      ottr:Triple(?subject,schema:endDate, ?end_date)
    } . 
    """

    feed_info_mut = feed_info.with_columns([
        (FEED_PUBLISHER_PREFIX + pl.col("feed_publisher_name")).alias("subject"),
        (pl.col("feed_start_date").str.strptime(pl.Date, "%Y-%m-%d")),
        (pl.col("feed_end_date").str.strptime(pl.Date, "%Y-%m-%d"))]
    ).collect()

    feed_info_mapping = mapping_prefixes + """
    t:Feed [?feed_publisher_name, xsd:anyURI ?feed_publisher_url, ?feed_lang, ?feed_start_date, ?feed_end_date, 
            ?feed_version,  xsd:anyURI ?subject ] :: {
      ottr:Triple(?subject,rdf:type, gtfs:Feed) ,
      ottr:Triple(?subject,dct:publisher,?feed_publisher_name) ,
      ottr:Triple(?subject,foaf:page,?feed_publisher_url) ,
      ottr:Triple(?subject,dct:language,?feed_lang) ,
      ottr:Triple(?subject,schema:startDate,?feed_start_date) ,
      ottr:Triple(?subject,schema:endDate,?feed_end_date) ,
      ottr:Triple(?subject,schema:version,?feed_version)
    } . 
    """

    shapes_mut = shapes.with_columns([
        (SHAPE_POINT_PREFIX + pl.col("shape_id") + "-" + pl.col("shape_pt_sequence")).alias("point"),
        (SHAPE_ID_PREFIX + pl.col("shape_id")).alias("shape_id")
    ]).select(["shape_id", "point"]).collect()

    shapes_mapping = mapping_prefixes + """
    t:Shapes [ xsd:anyURI ?shape_id, xsd:anyURI ?point] :: {
      ottr:Triple(?shape_id, rdf:type, gtfs:Shape) ,
      ottr:Triple(?shape_id, gtfs:shapePoint, ?point) 
      } . 
    """

    shape_points_mut = shapes.with_columns([
        (SHAPE_POINT_PREFIX + pl.col("shape_id") + "-" + pl.col("shape_pt_sequence")).alias("subject"),
        pl.col("shape_pt_lat").cast(pl.Float64),
        pl.col("shape_pt_lon").cast(pl.Float64)
    ]).drop(["shape_id"]).collect()

    shape_points_mapping = mapping_prefixes + """
    t:ShapePoints [?shape_pt_lat, ?shape_pt_lon, ?shape_pt_sequence, ?shape_dist_traveled,  xsd:anyURI ?subject ] :: {
      ottr:Triple(?subject, rdf:type, gtfs:ShapePoint) ,
      ottr:Triple(?subject,geo:lat,?shape_pt_lat) ,
      ottr:Triple(?subject,geo:long,?shape_pt_lon) ,
      ottr:Triple(?subject,gtfs:pointSequence,?shape_pt_sequence) ,
      ottr:Triple(?subject,gtfs:distanceTraveled,?shape_dist_traveled)
    } . 
    """

    frequencies_mut = frequencies.with_columns([
        (FREQUENCY_ID_PREFIX + pl.col("trip_id") + "-" + pl.col("start_time")).alias("subject"),
        (TRIP_ID_PREFIX + pl.col("trip_id")).alias("trip_id"),
        pl.col("exact_times").cast(pl.Boolean)
    ]).collect()

    frequencies_mapping = mapping_prefixes + """
    t:Frequencies [xsd:anyURI ?trip_id, ?start_time, ?end_time, ?headway_secs, ?exact_times,  xsd:anyURI ?subject ] :: {
      ottr:Triple(?subject,rdf:type, gtfs:Frequency) ,
      ottr:Triple(?subject,gtfs:trip,?trip_id) ,
      ottr:Triple(?subject,gtfs:startTime,?start_time) ,
      ottr:Triple(?subject,gtfs:endTime,?end_time) ,
      ottr:Triple(?subject,gtfs:headwaySeconds,?headway_secs) ,
      ottr:Triple(?subject,gtfs:exactTimes,?exact_times)
    } . 
    """

    m = Mapping([stop_times_mapping,
                 trips_mapping,
                 routes_mapping,
                 agency_mapping,
                 stops_mapping,
                 calendar_dates_mapping,
                 calendar_mapping,
                 feed_info_mapping,
                 shapes_mapping,
                 shape_points_mapping,
                 frequencies_mapping])

    m.expand("t:StopTimes", stop_times_mut, ["subject"])
    m.expand("t:Trips", trips_mut, ["trip_id"])
    m.expand("t:Routes", routes_mut, ["route_id"])
    m.expand("t:Agency", agency_mut, ["agency_id"])
    m.expand("t:Stops", stops_mut, ["stop_id"])
    m.expand("t:CalendarDateRules", calendar_dates_mut, ["subject"])
    m.expand("t:CalendarRules", calendar_mut, ["service_id"])
    m.expand("t:Feed", feed_info_mut, ["subject"])
    m.expand("t:Shapes", shapes_mut, ["point"])
    m.expand("t:ShapePoints", shape_points_mut, ["subject"])
    m.expand("t:Frequencies", frequencies_mut, ["subject"])

    time_end = time.time()
    print(f"Mapping took {round(time_end-time_start, 2)} seconds")
    return m


@pytest.mark.skip("dataset not yet in git")
def test_q15(mapping):
    query = """
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX foaf: <http://xmlns.com/foaf/0.1/>
PREFIX gtfs: <http://vocab.gtfs.org/terms#>
PREFIX geo: <http://www.w3.org/2003/01/geo/wgs84_pos#>
PREFIX dct: <http://purl.org/dc/terms/>

SELECT * WHERE { 
	?stop a gtfs:Stop .
	?stop ?p ?str .
	FILTER regex (?str, '000000000000000002')
}
"""
    print(mapping.query(query))

@pytest.mark.skip("dataset not yet in git")
def test_q7(mapping):
    query = """
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX foaf: <http://xmlns.com/foaf/0.1/>
PREFIX gtfs: <http://vocab.gtfs.org/terms#>
PREFIX geo: <http://www.w3.org/2003/01/geo/wgs84_pos#>
PREFIX dct: <http://purl.org/dc/terms/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
PREFIX gtfsroute: <http://transport.linkeddata.es/madrid/metro/routes/>
PREFIX gtfsaccessible: <http://transport.linkeddata.es/resource/WheelchairBoardingStatus/>

SELECT DISTINCT ?routeShortName ?routeDescription ?tripShortName ?stopDescription ?stopLat ?stopLong   
WHERE {
	?route a gtfs:Route .
	OPTIONAL { ?route gtfs:shortName ?routeShortName . }
	OPTIONAL { ?route dct:description ?routeDescription . }

	?trip a gtfs:Trip .
	OPTIONAL { ?trip gtfs:shortName ?tripShortName . }
	?trip gtfs:service ?service .
	?trip gtfs:route ?route .

	?stopTime a gtfs:StopTime . 
	?stopTime gtfs:trip ?trip . 
	?stopTime gtfs:stop ?stop . 

	?stop a gtfs:Stop . 
	OPTIONAL { ?stop dct:description ?stopDescription . }
	OPTIONAL {
		?stop geo:lat ?stopLat . 
		?stop geo:long ?stopLong . 
	}
	?stop gtfs:wheelchairAccessible gtfsaccessible:1 .
    FILTER (?route=gtfsroute:0000000000000000000c)
}
"""
    print(mapping.query(query))
@pytest.mark.skip("dataset not yet in git")
def test_queries(mapping):
    query_path = GTFS_PATH / "queries"
    errors = 0
    perf = []
    #10 is not relevant as the data type is wrong in the test data
    #7 is crashing with sigkill
    for fname in os.listdir(query_path):
        if fname in [f'q{str(i)}.rq' for i in [1,2,3,4,5,6,8,9,11,12,13,14,15,16,17,18]]:
            with open(query_path / fname) as f:
                q = f.read()
                print(fname.replace(".rq", "\n"))
                print(q)
                start = time.time()
                try:
                    res = mapping.query(q)
                    used_time = round(time.time()-start,2)
                    perf.append((fname, used_time))
                    print(f"Took {used_time} seconds")
                    print(f"Ok {res.height}")

                except:
                    print("Error")
                    errors+=1
                print("\n")


    print(f"N errors {errors}\n")
    print(f"Perf: {perf}")





import polars as pl
import pytest
from polars.testing import assert_frame_equal
pl.Config.set_fmt_str_lengths(300)


@pytest.fixture(scope="function")
def pizzas_mapping():
    from maplib import Mapping, Prefix, Template, Argument, Parameter, Variable, RDFType, Triple, a
    import polars as pl
    pl.Config.set_fmt_str_lengths(150)

    pi = "https://github.com/DataTreehouse/maplib/pizza#"
    df = pl.DataFrame({
        "p":[pi + "Hawaiian", pi + "Grandiosa"],
        "c":[pi + "CAN", pi + "NOR"],
        "ings": [[pi + "Pineapple", pi + "Ham"],
                 [pi + "Pepper", pi + "Meat"]]
    })
    #print(df)

    pi = Prefix("pi", pi)

    p_var = Variable("p")
    c_var = Variable("c")
    ings_var = Variable("ings")

    template = Template(
        iri= pi.suf("PizzaTemplate"),
        parameters= [
            Parameter(variable=p_var, rdf_type=RDFType.IRI()),
            Parameter(variable=c_var, rdf_type=RDFType.IRI()),
            Parameter(variable=ings_var, rdf_type=RDFType.Nested(RDFType.IRI()))
        ],
        instances= [
            Triple(p_var, a(), pi.suf("Pizza")),
            Triple(p_var, pi.suf("fromCountry"), c_var),
            Triple(p_var, pi.suf("hasIngredient"), Argument(term=ings_var, list_expand=True), list_expander="cross")
        ]
    )

    m = Mapping()
    m.expand(template, df)
    hpizzas = """
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>
    CONSTRUCT { ?p a pi:HeterodoxPizza } 
    WHERE {
        ?p a pi:Pizza .
        ?p pi:hasIngredient pi:Pineapple .
    }"""
    m.insert(hpizzas)
    return m


def test_simple_query_no_error(pizzas_mapping):
    res = pizzas_mapping.query("""
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>

    SELECT ?p WHERE {
    ?p a pi:HeterodoxPizza
    }
    """)

    expected_df = pl.DataFrame({"p": ["<https://github.com/DataTreehouse/maplib/pizza#Hawaiian>"]})
    assert_frame_equal(res, expected_df)


import polars as pl
from maplib import Mapping
from polars.testing import assert_frame_equal

def test_multi_filter_equals():
    m = Mapping([])
    df = m.query("""
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(?a = :hello2)
    }
    """)
    assert_frame_equal(df, pl.DataFrame({"a":["<http://example.net/hello2>"]}))

def test_multi_filter_equals_mirrored():
    m = Mapping([])
    df = m.query("""
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(:hello2 = ?a)
    }
    """)
    assert_frame_equal(df, pl.DataFrame({"a":["<http://example.net/hello2>"]}))

def test_multi_filter_equals_two_multi():
    m = Mapping([])
    df = m.query("""
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    {VALUES (?b) { (:hello2) }} UNION {VALUES (?b) { (1) }}
    FILTER(?b = ?a)
    }
    """)
    assert_frame_equal(df, pl.DataFrame({"a":["<http://example.net/hello2>"]}))


def test_multi_filter_is_in():
    m = Mapping([])
    df = m.query("""
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(?a in (:hello2, "a"))
    }
    """)
    assert_frame_equal(df, pl.DataFrame({"a":["<http://example.net/hello2>"]}))



from pathlib import Path
from typing import Union, List, Dict, Optional, Callable, Tuple, Literal as LiteralType
from polars import DataFrame
from datetime import datetime, date


class RDFType:
    """
    The type of a column containing a RDF variable.
    For instance, xsd:string is RDFType.Literal("http://www.w3.org/2001/XMLSchema#string")
    """
    IRI: Callable[[], "RDFType"]
    Blank: Callable[[], "RDFType"]
    Literal: Callable[[Union[str, "IRI"]], "RDFType"]
    Nested: Callable[["RDFType"], "RDFType"]
    Unknown: Callable[[], "RDFType"]

class Variable:
    """
    A variable in a template.
    """
    name:str

    def __init__(self, name: str):
        """
        Create a new variable.
        :param name: The name of the variable.
        """
        ...


class IRI:
    """
    An IRI.
    """

    def __init__(self, iri: str):
        """
        Create a new IRI
        :param iri: IRI (without < and >).
        """


class Prefix:
    """
    A prefix that can be used to ergonomically build iris.
    """

    def __init__(self, prefix, iri):
        """
        Create a new prefix.
        :param prefix: The name of the prefix
        :param iri: The prefix IRI.
        """

    def suf(self, suffix: str) -> IRI:
        """
        Create a IRI by appending the suffix.
        :param suffix: The suffix to append.
        :return:
        """

class Literal:
    """
    An RDF literal.
    """
    value: str
    data_type: Optional[IRI]
    language: Optional[str]

    def __init__(self, value: str, data_type: IRI = None, language: str = None):
        """
        Create a new RDF Literal
        :param value: The lexical representation of the value.
        :param data_type: The data type of the value (an IRI).
        :param language: The language tag of the value.
        """

    def to_native(self) -> Union[int, float, bool, str, datetime, date]:
        """

        :return:
        """


class Parameter:
    def __init__(self,
                 variable: Variable,
                 optional: bool = False,
                 allow_blank: bool = True,
                 rdf_type: RDFType = None):
        """
        Create a new parameter for a Template.
        :param variable: The variable.
        :param optional: Can the variable be unbound?
        :param allow_blank: Can the variable be bound to a blank node?
        :param rdf_type: The type of the variable. Can be nested.
        """


class Argument:
    def __init__(self, term: Union[Variable, IRI, Literal], list_expand: bool):
        """
        An argument for a template instance.
        :param term: The term.
        :param list_expand: Should the argument be expanded? Used with the list_expander argument of instance.
        """


class Instance:
    def __init__(self,
                 iri: IRI,
                 arguments: List[Union[Argument, Variable, IRI, Literal]],
                 list_expander: LiteralType["cross", "zipMin", "zipMax"] = None):
        """
        A template instance.
        :param iri: The IRI of the template to be instantiated.
        :param arguments: The arguments for template instantiation.
        :param list_expander: (How) should we do list expansion?
        """


class Template:
    """
    An OTTR Template.
    """

    def __init__(self,
                 iri: IRI,
                 parameters: List[Parameter],
                 instances: List[Instance]):
        """
        Create a new OTTR Template
        :param iri: The IRI of the template
        :param parameters:
        :param instances:
        """

    def instance(self,
                 arguments: List[Union[Argument, Variable, IRI, Literal]],
                 list_expander: LiteralType["cross", "zipMin", "zipMax"] = None) -> Instance:
        """

        :param arguments: The arguments to the template.
        :param list_expander: (How) should we list-expand?
        :return:
        """

def Triple(subject:Union["Argument", IRI, Variable],
           predicate:Union["Argument", IRI,Variable],
           object:Union["Argument", IRI, Variable, Literal],
           list_expander:Optional[LiteralType["cross", "zipMin", "zipMax"]]=None):
    """
    An OTTR Triple Pattern used for creating templates.
    This is the basis pattern which all template instances are rewritten into.
    Equivalent to:

    >>> ottr = Prefix("http://ns.ottr.xyz/0.4/")
    ... Instance(ottr.suf("Triple"), subject, predicate, object, list_expander)

    :param subject:
    :param predicate:
    :param object:
    :param list_expander:
    :return:
    """

class XSD():
    """
    The xsd namespace, for convenience.
    """
    boolean:IRI
    byte:IRI
    date:IRI
    dateTime:IRI
    dateTimeStamp:IRI
    decimal:IRI
    double:IRI
    duration:IRI
    float:IRI
    int_:IRI
    integer:IRI
    language:IRI
    long:IRI
    short:IRI
    string:IRI

    def __init__(self):
        """
        Create the xsd namespace helper.
        """

def a() -> IRI:
    """
    :return: IRI("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
    """

# END COMMON WITH CHRONTEXT

ParametersType = Dict[str, Tuple[DataFrame, Dict[str, RDFType]]]


class ValidationReport:
    """
    SHACL Validation report.
    Only constructed by maplib.
    """
    conforms: bool

    def results(self,
                native_dataframe: bool = False) -> Optional[DataFrame]:
        """
        Return the results of the validation report, if they exist.

        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :return: The SHACL validation report, as a DataFrame
        """

    def details(self,
                native_dataframe: bool = False) -> Optional[DataFrame]:
        """
        Returns the details of the validation report.
        Only available if validation was called with include_details=True.

        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :return: Details of the SHACL validation report, as a DataFrame
        """

    def graph(self) -> "Mapping":
        """
        Creates a new mapping object where the base graph is the validation report with results.
        Includes the details of the validation report in the new graph if they exist.

        :return:
        """


class Mapping:
    """
    A mapping session allowing:

    * Iterative mapping using OTTR templates
    * Interactive SPARQL querying and enrichment
    * SHACL validation

    Usage:

    >>> from maplib import Mapping
    ... doc = '''
    ... :prefix ex:<http://example.net/ns#>.
    ... ex:ExampleTemplate [?MyValue] :: {
    ...    ottr:Triple(ex:myObject, ex:hasValue, ?MyValue)
    ... } .'''
    ... m = Mapping(doc)

    :param documents: a stOTTR document or a list of these
    :param caching_folder: a folder to cache the triples as Parquet files
    """

    def __init__(self, documents: Union[str, List[str]] = None, caching_folder: str = None) -> Mapping: ...

    def add_template(self, template: "Template"):
        """
        Add a template to the mapping. Overwrites any existing template with the same IRI.
        :param template: The template to add.
        :return:
        """

    def expand(self, template: Union[str, "Template", IRI],
               df: DataFrame = None, unique_subset: List[str] = None) -> None:
        """
        Expand a template using a DataFrame
        Usage:

        >>> m.expand("ex:ExampleTemplate", df)
        ... m.expand("ex:ExampleTemplate", df, unique_subsets=["MyValue"])

        If the template has no arguments, the df argument is not necessary.

        :param template: Template, IRI, IRI string or prefixed template name.
        :param df: DataFrame where the columns have the same names as the template arguments
        :param unique_subset: DataFrame column names known to be unique e.g. ["colA", "colB"], for a performance boost (reduce costly deduplication)
        """

    def expand_default(self, df: DataFrame, primary_key_column: str, template_prefix: str = None,
                       predicate_uri_prefix: str = None) -> str:
        """
        Create a default template and expand it based on a dataframe.
        Usage:

        >>> template_string = m.expand_default(df, "myKeyCol")
        ... print(template_string)

        :param df: DataFrame where the columns have the same names as the template arguments
        :param primary_key_column: This column will be the subject of all triples in the generated template.
        :param template_prefix: Prefix of the template - the name is auto-generated.
        :param predicate_uri_prefix: Prefix of the predicates/verbs in the generated template, names are derived from column names.
        :return: The generated template
        """

    def query(self, query: str,
              parameters: ParametersType = None,
              include_datatypes:bool = False,
              native_dataframe:bool = False,
              graph: str = None) -> Union[
        DataFrame,
        Dict[str, Union[DataFrame, Dict[str, str]]],
        List[Union[DataFrame, Dict[str, Union[DataFrame, Dict[str, str]]]]],
        None]:
        """
        Query the contained knowledge graph using SPARQL
        Currently, SELECT, CONSTRUCT and INSERT are supported.
        Usage:

        >>> df = mapping.query('''
        ... PREFIX ex:<http://example.net/ns#>
        ... SELECT ?obj1 ?obj2 WHERE {
        ...    ?obj1 ex:hasObj ?obj2
        ... }''')
        ... print(df)

        :param query: The SPARQL query string
        :param parameters: PVALUES Parameters, a DataFrame containing the value bindings in the custom PVALUES construction.
        :param native_dataframe: Return columns with maplib-native formatting. Useful for round-trips.
        :param include_datatypes: Datatypes are not returned by default, set to true to return a dict with the solution mappings and the datatypes.
        :param graph: The IRI of the graph to query.
        :return: DataFrame (Select), list of DataFrames (Construct) containing results, or None for Insert-queries

        """

    def insert(self, query: str,
               parameters: ParametersType = None,
               transient: bool = False,
               source_graph: str = None,
               target_graph: str = None):
        """
        Insert the results of a Construct query in the graph.
        Useful for being able to use the same query for inspecting what will be inserted and actually inserting.
        Usage:

        >>> m = Mapping(doc)
        ... # Omitted
        ... hpizzas = '''
        ... PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        ... PREFIX ing:<https://github.com/magbak/maplib/pizza/ingredients#>
        ... CONSTRUCT { ?p a pizza:HeterodoxPizza }
        ... WHERE {
        ... ?p a pizza:Pizza .
        ... ?p pizza:hasIngredient ing:Pineapple .
        ... }'''
        ... m.insert(hpizzas)

        :param query: The SPARQL Insert query string
        :param parameters: PVALUES Parameters, a DataFrame containing the value bindings in the custom PVALUES construction.
        :param transient: Should the inserted triples be included in exports?
        :param source_graph: The IRI of the source graph to execute the construct query.
        :param target_graph: The IRI of the target graph to insert into.
        :return: None
        """

    def validate(self,
                 shape_graph: str,
                 include_details: bool = False,
                 include_conforms: bool = False,
                 ) -> ValidationReport:
        """
        Validate the contained knowledge graph using SHACL
        Assumes that the contained knowledge graph also contains SHACL Shapes.

        :param shape_graph: The IRI of the Shape Graph.
        :param include_details: Include details of SHACL evaluation alongside the report. Currently uses a lot of memory.
        :param include_conforms: Include those results that conformed. Also applies to details.
        :return: Validation report containing a report (report.df) and whether the graph conforms (report.conforms)
        """

    def read_triples(self,
                     file_path: Union[str, Path],
                     format: LiteralType["ntriples", "turtle", "rdf/xml", "xml", "rdfxml"] = None,
                     base_iri: str = None,
                     transient: bool = False,
                     parallel: bool = False,
                     checked: bool = True,
                     deduplicate: bool = True,
                     graph: str = None) -> None:
        """
        Reads triples from a file path.
        You can specify the format, or it will be derived using file extension, e.g. filename.ttl or filename.nt.
        Specify transient if you only want the triples to be available for further querying and validation,
        but not persisted using write-methods.

        Usage:

        >>> m.read_triples("my_triples.ttl")

        :param file_path: The path of the file containing triples
        :param format: One of "ntriples", "turtle", "rdf/xml", otherwise it is inferred from the file extension.
        :param base_iri: Base iri
        :param transient: Should these triples be included when writing the graph to the file system?
        :param parallel: Parse triples in parallel, currently only NTRiples. Assumes all prefixes are in the beginning of the document.
        :param checked: Check IRIs etc.
        :param deduplicate: Set to true by default, disable to increase throughput for large files containing only unique triples.
        :param graph: The IRI of the graph to read the triples into.
        """

    def read_triples_string(self,
                            s: str,
                            format: LiteralType["ntriples", "turtle", "rdf/xml", "xml", "rdfxml"],
                            base_iri: str = None,
                            transient: bool = False,
                            parallel: bool = False,
                            checked: bool = True,
                            deduplicate: bool = True,
                            graph: str = None) -> None:
        """
        Reads triples from a string.
        Specify transient if you only want the triples to be available for further querying and validation,
        but not persisted using write-methods.

        Usage:

        >>> m.read_triples(my_ntriples_string, format="ntriples")

        :param s: String containing serialized triples.
        :param format: One of "ntriples", "turtle", "rdf/xml".
        :param base_iri: Base iri
        :param transient: Should these triples be included when writing the graph to the file system?
        :param parallel: Parse triples in parallel, currently only NTRiples. Assumes all prefixes are in the beginning of the document.
        :param checked: Check IRIs etc.
        :param deduplicate: Set to true by default, disable to increase throughput for large files containing only unique triples.
        :param graph: The IRI of the graph to read the triples into.
        """

    def write_ntriples(self, file_path: Union[str, Path], graph: str = None) -> None:
        """
        Write the non-transient triples to the file path specified in the NTriples format.

        Usage:

        >>> m.write_ntriples("my_triples.nt")

        :param file_path: The path of the file containing triples
        :param graph: The IRI of the graph to write.
        """

    def write_ntriples_string(self, graph: str = None) -> str:
        """
        Write the non-transient triples to a string in memory.

        Usage:

        >>> s = m.write_ntriples_string()

        :param graph: The IRI of the graph to write.
        :return Triples in mapping in the NTriples format (potentially a large string)
        """

    def write_native_parquet(self, folder_path: Union[str, Path], graph: str = None) -> None:
        """
        Write non-transient triples using the internal native Parquet format.

        Usage:

        >>> m.write_native_parquet("output_folder")

        :param folder_path: The path of the folder to write triples in the native format.
        :param graph: The IRI of the graph to write.
        """

    def create_sprout(self):
        """
        A sprout is a simplified way of dealing with multiple graphs.
        See also `maplib.maplib.Mapping.insert_sprout` and `maplib.maplib.Mapping.detach_sprout`

        :return:
        """

    def insert_sprout(self, query: str,
                      parameters: ParametersType = None,
                      transient: bool = False,
                      source_graph: str = None,
                      target_graph: str = None):
        """
        Insert the results of a Construct query in a sprouted graph, which is created if no sprout is active.
        Sprouts are simplified way of dealing with multiple graphs.
        Useful for being able to use the same query for inspecting what will be inserted and actually inserting.
        See also `maplib.maplib.Mapping.detach_sprout`

        Usage:

        >>> m = Mapping(doc)
        ... m.create_sprout()
        ... # Omitted
        ... hpizzas = '''
        ... PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
        ... PREFIX ing:<https://github.com/magbak/maplib/pizza/ingredients#>
        ... CONSTRUCT { ?p a pizza:HeterodoxPizza }
        ... WHERE {
        ... ?p a pizza:Pizza .
        ... ?p pizza:hasIngredient ing:Pineapple .
        ... }'''
        ... m.insert_sprout(hpizzas)

        :param query: The SPARQL Insert query string
        :param parameters: PVALUES Parameters, a DataFrame containing the value bindings in the custom PVALUES construction.
        :param transient: Should the inserted triples be included in exports?
        :param source_graph: The IRI of the source graph to execute the construct query.
        :param target_graph: The IRI of the target graph to insert into.
        :return: None
        """

    def detach_sprout(self) -> Mapping:
        """
        Detaches and returns the sprout from the mapping.

        @return: The sprout as its own Mapping.
        """


import polars as pl
from maplib import Mapping, Template, IRI, Prefix, Triple, Variable, Parameter
from polars.testing import assert_frame_equal

def test_create_mapping_from_empty_polars_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyValue] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyValue)
    } .
    """

    df = pl.DataFrame({"MyValue": []})
    mapping = Mapping([doc])
    mapping.expand("http://example.net/ns#ExampleTemplate", df)

def test_create_mapping_with_optional_value_missing_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyValue, ??MyOtherValue] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyValue),
    ottr:Triple(ex:myObject, ex:hasOtherValue, ?MyOtherValue)
    } .
    """

    df = pl.DataFrame({"MyValue": ["A"]})
    mapping = Mapping([doc])
    mapping.expand("http://example.net/ns#ExampleTemplate", df)
    qres = mapping.query(
    """
    PREFIX ex:<http://example.net/ns#>
    
    SELECT ?A WHERE {
    ?obj1 ex:hasValue ?A
    } 
    """
    )
    expected_df = pl.DataFrame({"A":["A"],})
    assert_frame_equal(qres, expected_df)

    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A WHERE {
        ?obj1 ex:hasOtherValue ?A
        } 
        """
    )
    assert qres.height == 0


def test_create_programmatic_mapping_with_optional_value_missing_df():
    df = pl.DataFrame({"MyValue": ["A"]})
    mapping = Mapping()
    ex = Prefix("ex","http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value), Parameter(my_other_value, optional=True)],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
            Triple(my_object, ex.suf("hasOtherValue"), my_other_value)
         ]
    )
    mapping.expand(template, df)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A WHERE {
        ?obj1 ex:hasValue ?A
        } 
        """
    )
    expected_df = pl.DataFrame({"A":["A"],})
    assert_frame_equal(qres, expected_df)

    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>
        
        SELECT ?A WHERE {
        ?obj1 ex:hasOtherValue ?A
        } 
        """
    )
    assert qres.height == 0

def test_create_mapping_from_empty_signature():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [] :: {
    ottr:Triple(ex:myObject, ex:hasObj, ex:myOtherObject)
    } .
    """

    mapping = Mapping([doc])
    mapping.expand("http://example.net/ns#ExampleTemplate")
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj1 ?obj2 WHERE {
        ?obj1 ex:hasObj ?obj2
        } 
        """
    )
    expected_df = pl.DataFrame({"obj1":"<http://example.net/ns#myObject>",
                                "obj2":"<http://example.net/ns#myOtherObject>"})

    assert_frame_equal(qres, expected_df)


def test_uri_subject_query():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [] :: {
    ottr:Triple(ex:myObject, ex:hasObj, ex:myOtherObject)
    } .
    """

    mapping = Mapping([doc])
    mapping.expand("http://example.net/ns#ExampleTemplate")
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj2 WHERE {
        ex:myObject ex:hasObj ?obj2
        } 
        """
    )
    expected_df = pl.DataFrame({"obj2":"<http://example.net/ns#myOtherObject>"})

    assert_frame_equal(qres, expected_df)


def test_programmatic_mapping():
    example_template = Template(
        iri=IRI("http://example.net/ns#ExampleTemplate"),
        parameters=[],
        instances=[Triple(
            IRI("http://example.net/ns#myObject"),
            IRI("http://example.net/ns#hasObj"),
            IRI("http://example.net/ns#myOtherObject"),
        )]
    )

    mapping = Mapping()
    mapping.add_template(example_template)
    mapping.expand("http://example.net/ns#ExampleTemplate")
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj2 WHERE {
        ex:myObject ex:hasObj ?obj2
        } 
        """
    )
    expected_df = pl.DataFrame({"obj2":"<http://example.net/ns#myOtherObject>"})

    assert_frame_equal(qres, expected_df)


def test_programmatic_mapping_to_string():
    ex = Prefix("ex", "http://example.net/ns#")
    example_template = Template(
        iri=ex.suf("ExampleTemplate"),
        parameters=[],
        instances=[Triple(
            ex.suf("myObject"),
            ex.suf("hasObj"),
            ex.suf("myOtherObject"),
        )]
    )
    assert str(example_template) == """<http://example.net/ns#ExampleTemplate> [ ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#myObject>,<http://example.net/ns#hasObj>,<http://example.net/ns#myOtherObject>)
} . 
"""


def test_programmatic_mapping_with_prefix():
    ex = Prefix("ex", "http://example.net/ns#")
    example_template = Template(
        iri=ex.suf("ExampleTemplate"),
        parameters=[],
        instances=[Triple(
            ex.suf("myObject"),
            ex.suf("hasObj"),
            ex.suf("myOtherObject"),
        )]
    )

    mapping = Mapping()
    mapping.expand(example_template)
    qres = mapping.query(
        """
        PREFIX ex:<http://example.net/ns#>

        SELECT ?obj2 WHERE {
        ex:myObject ex:hasObj ?obj2
        } 
        """
    )
    expected_df = pl.DataFrame({"obj2":"<http://example.net/ns#myOtherObject>"})

    assert_frame_equal(qres, expected_df)


def test_create_and_write_no_bug_string_lit_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyString] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyString)
    } .
    """

    df = pl.DataFrame({"MyString": ["\"!!\n#\"", "ABC#123\n\t\r"]})
    mapping = Mapping([doc])
    mapping.expand("http://example.net/ns#ExampleTemplate", df)
    ntfile = "create_and_write_no_bug_string_lit.nt"
    mapping.write_ntriples(ntfile)
    with open(ntfile) as f:
        lines = f.readlines()

    lines.sort()
    assert lines == ['<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "ABC#123\\n\\t\\r" .\n',
                     '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "\\"!!\\n#\\"" .\n']

def test_create_and_write_no_bug_bool_lit_df():
    doc = """
    @prefix ex:<http://example.net/ns#>.
    ex:ExampleTemplate [?MyBool] :: {
    ottr:Triple(ex:myObject, ex:hasValue, ?MyBool)
    } .
    """

    df = pl.DataFrame({"MyBool": [True, False]})
    mapping = Mapping([doc])
    mapping.expand("http://example.net/ns#ExampleTemplate", df)
    ntfile = "create_and_write_no_bug_bool_lit.nt"
    mapping.write_ntriples(ntfile)
    with open(ntfile) as f:
        lines = f.readlines()

    lines.sort()
    assert lines == ['<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "false"^^<http://www.w3.org/2001/XMLSchema#boolean> .\n',
                     '<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "true"^^<http://www.w3.org/2001/XMLSchema#boolean> .\n']


def test_create_and_write_no_bug_lang_string_lit_df():
    mapping = Mapping()
    mapping.read_triples_string('<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "HELLO!!"@en .', format="ntriples")
    ntfile = "create_and_write_no_bug_string_lit.nt"
    mapping.write_ntriples(ntfile)
    with open(ntfile) as f:
        lines = f.readlines()

    lines.sort()
    assert lines == ['<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "HELLO!!"@en .\n']



import polars as pl
import pytest
import rdflib
from polars.testing import assert_frame_equal
import pathlib
from maplib import Mapping

pl.Config.set_fmt_str_lengths(300)


PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"

@pytest.fixture(scope="function")
def blank_person_mapping():
    # The following example comes from https://primer.ottr.xyz/01-basics.html

    doc = """
    @prefix rdf: 	<http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
    @prefix rdfs: 	<http://www.w3.org/2000/01/rdf-schema#> . 
    @prefix owl: 	<http://www.w3.org/2002/07/owl#> . 
    @prefix xsd: 	<http://www.w3.org/2001/XMLSchema#> . 
    @prefix foaf: 	<http://xmlns.com/foaf/0.1/> . 
    @prefix dbp: 	<http://dbpedia.org/ontology/> . 
    @prefix ex: 	<http://example.com/ns#> . 
    @prefix ottr: 	<http://ns.ottr.xyz/0.4/> . 
    @prefix ax: 	<http://tpl.ottr.xyz/owl/axiom/0.1/> . 
    @prefix rstr: 	<http://tpl.ottr.xyz/owl/restriction/0.1/> .
    ex:Person[ ?firstName, ?lastName, ottr:IRI ?email ] :: {
      ottr:Triple(_:person, rdf:type, foaf:Person ),
      ottr:Triple(_:person, foaf:firstName, ?firstName ),
      ottr:Triple(_:person, foaf:lastName, ?lastName ),
      ottr:Triple(_:person, foaf:mbox, ?email )
    } .
    """
    m = Mapping([doc])
    df = pl.DataFrame({"firstName": ["Ann", "Bob"],
                       "lastName": ["Strong", "Brite"],
                       "email": ["mailto:ann.strong@example.com", "mailto:bob.brite@example.com"]})
    m.expand("ex:Person", df)
    return m


def test_simple_query_no_error(blank_person_mapping):
    qres = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?firstName ?lastName WHERE {
        ?p a foaf:Person .
        ?p foaf:lastName ?lastName .
        ?p foaf:firstName ?firstName .
        } ORDER BY ?firstName ?lastName
        """).sort(["firstName", "lastName"])
    #Todo: Fix multitype sorting
    expected_df = pl.DataFrame({"firstName": ["Ann", "Bob"],
                                "lastName": ["Strong", "Brite"]})

    assert_frame_equal(qres, expected_df)


def test_simple_query_blank_node_output_no_error(blank_person_mapping):
    blank_person_mapping.write_ntriples("out.nt")
    gr = rdflib.Graph()
    gr.parse("out.nt", format="ntriples")
    res = gr.query("""
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?firstName ?lastName WHERE {
            ?p a foaf:Person .
            ?p foaf:lastName ?lastName .
            ?p foaf:firstName ?firstName .
            } ORDER BY ?firstName ?lastName
            """)
    assert len(res) == 2


def test_multi_datatype_query_no_error(blank_person_mapping):
    res = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?v ?o WHERE {
        ?s ?v ?o .
        } 
        """)
    by = ["s","v","o"]
    df = res.sort(by=by)
    filename = TESTDATA_PATH / "multi_datatype_query.csv"
    #df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    assert_frame_equal(df, expected_df)


def test_multi_datatype_union_query_no_error(blank_person_mapping):
    res = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?o WHERE {
        {?s foaf:firstName ?o .}
        UNION {
        ?s a ?o .
        }
        } 
        """)
    by = ["s","o"]
    df = res.sort(by=by)
    filename = TESTDATA_PATH / "multi_datatype_union_query.csv"
    #df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    assert_frame_equal(df, expected_df)


def test_multi_datatype_union_sort_query(blank_person_mapping):
    df = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?o WHERE {
        {?s foaf:firstName ?o .}
        UNION {
        ?s a ?o .
        }
        } ORDER BY ?o ?s
        """)
    filename = TESTDATA_PATH / "multi_datatype_union_sort_query.csv"
    #df.write_csv(filename)
    expected_df = pl.scan_csv(filename).collect()
    assert_frame_equal(df, expected_df)


def test_multi_datatype_union_sort_desc1_query(blank_person_mapping):
    df = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?o WHERE {
        {?s foaf:firstName ?o .}
        UNION {
        ?s a ?o .
        }
        } ORDER BY DESC(?o) ?s
        """)
    filename = TESTDATA_PATH / "multi_datatype_union_sort_desc1_query.csv"
    #df.write_csv(filename)
    expected_df = pl.scan_csv(filename).collect()
    assert_frame_equal(df, expected_df)

def test_multi_datatype_union_query_native_df(blank_person_mapping):
    res = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?o WHERE {
        {?s foaf:firstName ?o .}
        UNION {
        ?s a ?o .
        }
        } 
        """, native_dataframe=True)
    by = ["s","o"]
    df = res.sort(by=by)
    filename = TESTDATA_PATH / "multi_datatype_union_query_native_df.parquet"
    #df.write_parquet(filename)
    expected_df = pl.scan_parquet(filename).sort(by).collect()
    assert_frame_equal(df, expected_df)

def test_multi_datatype_left_join_query_no_error(blank_person_mapping):
    res = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?o WHERE {
        ?s foaf:firstName ?o .
        OPTIONAL {
        ?s a ?o .
        }
        } 
        """)
    by = ["s","o"]
    df = res.sort(by=by)
    filename = TESTDATA_PATH / "multi_datatype_leftjoin_query.csv"
    #df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    assert_frame_equal(df, expected_df)

#This test is skipped due to a bug in Polars.
def test_multi_datatype_join_query_two_vars_no_error(blank_person_mapping):
    res = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?o WHERE {
        ?s foaf:firstName ?o .
        {
        {?s foaf:firstName ?o }
        UNION
        {?s a ?o }.
        }
        } 
        """)
    by = ["s","o"]
    df = res.sort(by=by)
    filename = TESTDATA_PATH / "multi_datatype_join_query_two_vars.csv"
    #df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    assert_frame_equal(df, expected_df)

def test_multi_datatype_join_query_no_error(blank_person_mapping):
    res = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s1 ?s2 ?o WHERE {
        ?s1 foaf:firstName ?o .
        {
        {?s2 foaf:firstName ?o }
        UNION
        {?s2 a ?o }.
        }
        } 
        """)
    by = ["s1", "s2","o"]
    df = res.sort(by=by)
    filename = TESTDATA_PATH / "multi_datatype_join_query.csv"
    #df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    assert_frame_equal(df, expected_df)


def test_multi_datatype_query_sorting_sorting(blank_person_mapping):
    res = blank_person_mapping.query("""
        PREFIX foaf:<http://xmlns.com/foaf/0.1/>

        SELECT ?s ?v ?o WHERE {
        ?s ?v ?o .
        } ORDER BY ?s ?v ?o
        """).sort(["s", "v", "o"])
    #TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "multi_datatype_query_sorting.csv"
    #res.write_csv(filename)
    expected_df = pl.scan_csv(filename).collect()
    assert_frame_equal(res, expected_df)


import polars as pl
import pytest
from polars.testing import assert_frame_equal

from maplib import Mapping, RDFType

pl.Config.set_fmt_str_lengths(300)


@pytest.fixture(scope="function")
def pizzas_mapping():
    doc = """
    @prefix pizza:<https://github.com/magbak/maplib/pizza#>.
    @prefix xsd:<http://www.w3.org/2001/XMLSchema#>.
    @prefix ex:<https://github.com/magbak/maplib/pizza#>.

    ex:Pizza[?p, xsd:anyURI ?c, List<ottr:IRI> ?is] :: {
    ottr:Triple(?p, a, pizza:Pizza),
    ottr:Triple(?p, pizza:fromCountry, ?c),
    cross | ottr:Triple(?p, pizza:hasIngredient, ++?is)
    }.
    """

    m = Mapping([doc])

    co = "https://github.com/magbak/maplib/countries#"
    pi = "https://github.com/magbak/maplib/pizza#"
    ing = "https://github.com/magbak/maplib/pizza/ingredients#"

    df = pl.DataFrame({"p": [pi + "Hawaiian", pi + "Grandiosa"],
                       "c": [co + "CAN", co + "NOR"],
                       "is": [[ing + "Pineapple", ing + "Ham"],
                              [ing + "Pepper", ing + "Meat"]]})
    m.expand("ex:Pizza", df)
    hpizzas = """
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>
    PREFIX ing:<https://github.com/magbak/maplib/pizza/ingredients#>
    CONSTRUCT { ?p a pizza:HeterodoxPizza } 
    WHERE {
        ?p a pizza:Pizza .
        ?p pizza:hasIngredient ing:Pineapple .
    }"""
    m.insert(hpizzas)
    return m


def test_simple_query_no_error(pizzas_mapping):
    res = pizzas_mapping.query("""
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>

    SELECT ?p WHERE {
    ?p a pizza:HeterodoxPizza
    }
    """)

    expected_df = pl.DataFrame({"p": ["<https://github.com/magbak/maplib/pizza#Hawaiian>"]})
    assert_frame_equal(res, expected_df)


def test_construct_pvalues(pizzas_mapping):
    h_df = pl.DataFrame({"h1": ["https://github.com/magbak/maplib/pizza#Hawaiian1", "https://github.com/magbak/maplib/pizza#Hawaiian2"],
                         "h2": ["https://github.com/magbak/maplib/pizza#Hawaiian3", "https://github.com/magbak/maplib/pizza#Hawaiian4"]})

    res = pizzas_mapping.query("""
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>

    CONSTRUCT {
        ?h1 a pizza:Pizza . 
        ?h2 a pizza:Pizza .
    } WHERE {
        PVALUES (?h1 ?h2) h
    }
    """, parameters={"h":(h_df, {"h1":RDFType.IRI(), "h2":RDFType.IRI()})})
    res0 = res[0]
    expected_dtypes = {'object': 'IRI', 'subject': 'IRI'}
    #assert res0.rdf_datatypes == expected_dtypes
    res1 = res[1]
    #assert res1.rdf_datatypes == expected_dtypes
    assert res0.height == 2
    assert res1.height == 2


def test_construct_pvalues2(pizzas_mapping):
    h_df = pl.DataFrame({"h1": ["https://github.com/magbak/maplib/pizza#Hawaiian", "https://github.com/magbak/maplib/pizza#Hawaiian2"]})
    res = pizzas_mapping.query("""
    PREFIX pizza:<https://github.com/magbak/maplib/pizza#>

    CONSTRUCT {
        ?h1 a pizza:Pizza . 
        ?i a pizza:Pizza .
    } WHERE {
        ?h1 pizza:hasIngredient ?i .
        PVALUES (?h1) h
    }
    """, parameters={"h":(h_df,{"h1":RDFType.IRI()})})
    res0 = res[0]
    expected_dtypes = {'object': 'IRI', 'subject': 'IRI'}
    #assert res0.rdf_datatypes == expected_dtypes
    res1 = res[1]
    #assert res1.rdf_datatypes == expected_dtypes
    assert res0.height == 1
    assert res1.height == 2
